% GET_DESCRIPTOR
%   GET_DESCRIPTOR returns the descriptor of a given line based on the gradient information of the image and orientation of the line
%   DESCRIPTOR = GET_DESCRIPTOR(PARAMETERS , GRADIENT, ANGLE)
%   PARAMETERS is the parameters needed for extracting the lines, see the main program for an explication
%   GRADIENT is the gradient image extracted by function VERTICAL_FEATURES
%   ANGLE is the orientation of the line (needed to make the descriptor rotation invariant
%   DES is a M element vector, where M is defined by M = 2 x #circular_areas x #histogram_bins (see [1] equation (5) for an explication
%
% Author: Davide Scaramuzza, June 2008
% Email: davide.scaramuzza@ieee.org
% If you use this code or re-implement the algorithm by yoursefl, please
% cite the publications:
%
% [1] Scaramuzza, D., Martinelli, A., and Siegwart, R. (2009)
% A Robust Descriptor for Tracking Vertical Lines in Omnidirectional Images and its Use in Mobile Robotics, 
% International Journal on Robotics Research, vol. 28, issue 2, February, 2009.
%
% [2] Scaramuzza, D., Criblez, N., Martinelli, A. and Siegwart, R. (2007), 
% Robust Feature Extraction and Matching for Omnidirectional Images, 
% Proceedings of the 6th International Conference on Field and Service Robotics, July, 2007

function descriptor = get_descriptor( parameters , Gradient, Angle)


%%% --------------------------------------------------------------------------------------------------- %%%
%%% --------------------------------------- initialisation -------------------------------------------- %%%
%%% --------------------------------------------------------------------------------------------------- %%%

%Pat_Window = parameters.Pat_Window;
Pat_Window = floor((parameters.Rmax-parameters.Rmin)/parameters.number_of_pattern_for_one_feat);
if mod(Pat_Window,2) == 0
    Pat_Window = Pat_Window + 1;
end
parameters.Pat_Window = Pat_Window;
number_of_pattern_for_one_feat = parameters.number_of_pattern_for_one_feat;
%_____________________________________________________________________________________________________________________________________________________

% image information
[height,width]=size(Gradient.mag);
centre_x = parameters.Centre.X_centre;                                             % coordonnee X du centre de l'image
centre_y = parameters.Centre.Y_centre;                                             % coordonnee Y du centre de l'image
%_____________________________________________________________________________________________________________________________________________________

% information about the ponderation gausian window
sigma = 1/6 * Pat_Window;
Gauss_Window = fspecial('gaussian',Pat_Window,sigma);
%_____________________________________________________________________________________________________________________________________________________


%%% --------------------------------------------------------------------------------------------------- %%%
%%% --------------------------------------------- MAIN ------------------------------------------------ %%%
%%% --------------------------------------------------------------------------------------------------- %%%

  
%0; find all the gradient (direction and magnitude) that we have in the all pattern (in the current vertical feature).
Feat = Find_gradient_along_vertical_features(Angle,Gauss_Window,Gradient,parameters);

%2; Construction of the descriptor
Descriptor = Construct_The_Descriptor_V2(Angle, Feat,parameters);

%3; save the information in the "Image_features" structure
descriptor = [];
for i=1:number_of_pattern_for_one_feat
    descriptor = [descriptor;
                  Descriptor(i).Descriptor.Descriptor];
end      

return

%--------------------------------------------------------------------------
%--------------------------------------------------------------------------



function [Feat] = Find_gradient_along_vertical_features(Angle, Gauss_Window, Gradient, parameters);

%_____________________________________________________________________________________________________________________________________________________
% Description of I/O
% Input
% Angle; it's the angle of the current vertical feature
% Gauss_Window; It's a ponderation gauss window (to transforme a square in a circle)
% centre_x; Centre of the image along the X direction
% centre_y; Centre of the image along the Y direction
% Gradient; It's a structure who contains all information about the direction and magnitude of the gradient of the present image
% Pat_Window;  size of the square window of pattern (in pixel)
% number_of_pattern_for_one_feat; number of pattern we take along a vertical feature
% N_Brother_patt_for_one_feat; number of brother patern we take along a features (it's the same pattern but juste some pixel dephase)

% Output
% Feat: It's a structure who contains all information about the gradient for each pattern we have along one vertical feature
%_____________________________________________________________________________________________________________________________________________________


%%% --------------------------------------------------------------------------------------------------- %%%
%%% --------------------------------------- initialisation -------------------------------------------- %%%
%%% --------------------------------------------------------------------------------------------------- %%%

% init
Pat_Window = parameters.Pat_Window;
number_of_pattern_for_one_feat = parameters.number_of_pattern_for_one_feat;
centre_x = parameters.Centre.X_centre;                                             % coordonnee X du centre de l'image
centre_y = parameters.Centre.Y_centre;  
Rmin = parameters.Rmin;
Rmax = parameters.Rmax;

LL=[];



Delta = Rmax-Rmin;

Dist_between_two_pat = 4;               % distance between two succesive brother pattern alog the same features, in pixel

if number_of_pattern_for_one_feat==1
    dist_inter_pat=0;
else
    dist_inter_pat = Delta/(number_of_pattern_for_one_feat-1);           % distance between two pattern
end

for z=1:number_of_pattern_for_one_feat
    LL=[LL  Rmin+(z-1)*dist_inter_pat];                                  % midle point of the vertical line
end


%%% --------------------------------------------------------------------------------------------------- %%%
%%% --------------------------------------------- Algo ------------------------------------------------ %%%
%%% --------------------------------------------------------------------------------------------------- %%%

LL=[];
pos = Delta/number_of_pattern_for_one_feat/2;
for z=1:number_of_pattern_for_one_feat
    LL=[LL  Rmin+(2*(z-1)+1)*pos];                                  % midle point of the vertical line
end

for q=1:length(LL)          % one loop per pattern
    
    L=LL(q)-floor(1/2)-Dist_between_two_pat;
    
    for i=1:1         % one loop more for each brother pattern (it's the same pattern but just dephase about some pixel)
           
        L=L+Dist_between_two_pat;
        
        Pos_centre_pat =  [round(centre_x - L*cos(Angle)) ; round(centre_y - L*sin(Angle))];   % position of the centre of the pattern
        
        demi_win = floor(Pat_Window/2);
        born_x_inf = Pos_centre_pat(1) - demi_win;
        born_x_sup = Pos_centre_pat(1) + demi_win;
        
        born_y_inf = Pos_centre_pat(2) - demi_win;
        born_y_sup = Pos_centre_pat(2) + demi_win;
        
        Dir_grad = Gradient.phase(born_x_inf:born_x_sup , born_y_inf:born_y_sup);                                  % direction of the gradient
        Mag_grad = Gradient.mag(born_x_inf:born_x_sup , born_y_inf:born_y_sup);                                   % magnitude of the gradient
        
        %_____________________________________________________________________________________________________________________________________________________
%        plot (just to verify)
%         hold on
%         plot (Pos_centre_pat(2),Pos_centre_pat(1),'o','color','r' );    
%         
%         plot([born_y_inf born_y_inf],[born_x_inf  born_x_sup],'color','r');         % gauche
%         plot([born_y_sup born_y_sup],[born_x_inf  born_x_sup],'color','r');         % droite
%         
%         plot([born_y_inf born_y_sup],[born_x_inf  born_x_inf],'color','r');         % haut
%         plot([born_y_inf born_y_sup],[born_x_sup  born_x_sup],'color','r');         % bas
%         hold off
        %_____________________________________________________________________________________________________________________________________________________
        
        
        Feat(i).pat(q).Grad_Direction = Dir_grad;
        Feat(i).pat(q).Grad_Magnitude = Mag_grad.*Gauss_Window; 
        
    end
end

return

%--------------------------------------------------------------------------
%--------------------------------------------------------------------------


function [Pat] = Construct_The_Descriptor_V2(Angle, Feat, parameters); 

number_of_pattern_for_one_feat = parameters.number_of_pattern_for_one_feat;
N_bins = parameters.N_bins;                % number of bins for the histogram

%_____________________________________________________________________________________________________________________________________________________
% Description of I/O
% Input
% Angle; it's the angle of the current vertical feature
% Feat: It's a structure who contains all information about the gradient for each pattern we have along one vertical feature
% number_of_pattern_for_one_feat; number of pattern we take along a
% vertical feature

% Output
% Pat: It's a structure who contains the information about the descriptor
%_____________________________________________________________________________________________________________________________________________________


%%% --------------------------------------------------------------------------------------------------- %%%
%%% --------------------------------------- initialisation -------------------------------------------- %%%
%%% --------------------------------------------------------------------------------------------------- %%%

% descriptor information
D=[];
N_cut = 2;              % number of the part of the descriptor
Inter_cut = (2*pi)/N_cut;

Extrema_cut = -pi;        % We work between -pi and pi

cut = [];

% Histo bins for the number of part that have a descriptor
for q=1:N_cut
    cut = [(-Extrema_cut + (q-1)*(2*Extrema_cut/N_cut))    cut];
end


% histogram information

Inter_bins = (2*pi)/N_bins;

Extrema_Histo = -pi;        % We work between -pi and pi

Bins_vector = [];

% Histo bins for each descriptor
for k=1:N_bins
    Bins_vector = [(-Extrema_Histo + (k-1)*(2*Extrema_Histo/N_bins))    Bins_vector];
end


% pattern information
info_grad = Feat(1).pat(1).Grad_Direction ;

longeur = length(info_grad(:,1));
largeur = length(info_grad(1,:));

centre_pat = ceil(longeur/2);              % the pattern is square

%%% --------------------------------------------------------------------------------------------------- %%%
%%% --------------------------------------------- Algo ------------------------------------------------ %%%
%%% --------------------------------------------------------------------------------------------------- %%%
for w=1:number_of_pattern_for_one_feat                       % becaus we have n pattern per vertical feature
    for k=1:length(Feat)
        
        % init of all histo
        for m=1:N_cut
            Histo(1:N_bins,m)=0;
        end
        %%%%%%
        
        
        % pattern information 
        Dir_grad = Feat(k).pat(w).Grad_Direction ;
        Mag_grad = Feat(k).pat(w).Grad_Magnitude ; 
        
        %_____________________________________________________________________________________________________________________________________________________
        
        % construct the histogram for the part we want
        for i=1:longeur
            for j=1:largeur
                
                Dir = Modul_2pi(Dir_grad(i,j)-Angle);
                Mag = Mag_grad(i,j);
                
                % find in wich histo goes this point
                pos_x = centre_pat -i;
                pos_y = centre_pat -j;
                
                gamma = atan2(pos_y,pos_x);                 % angle of the point with the same ref than the camera
                
                gamma =  Modul_2pi(gamma - Angle);                     % angle of the point with the ref of the pattern
                
                plus_b = find(cut>=gamma);
                ind = min(plus_b);                         % index of the right histo for the current point
                %%%%%%
                
                All_bins_minus = find(Bins_vector<=Dir);
                minus_bins = max(All_bins_minus);                  % floor neighbourg
                
                All_bins_plus = find(Bins_vector>=Dir);
                plus_bins = min(All_bins_plus);                     % ceil neighbourg
                
                if minus_bins==plus_bins           % the direction of the gradient is right on the bins
                    
                    Histo(minus_bins,ind) =  Histo(minus_bins,ind) + Mag;         % on ajoute mag pour tenir compte de la magnitude dans l'histograme
                    
                else
                    if isempty(minus_bins)       % condition si notre direction est dans les bins exterieur
                        factor_minus = (Bins_vector(plus_bins)-Dir) / Inter_bins;                     % quanty factor to put in the minus bins
                        factor_plus = (Dir-(-pi)) / Inter_bins;                      % quanty factor to put in the plus bins
                        
                        Histo(end,ind) =  Histo(end,ind) + factor_minus*Mag;                % because -pi = +pi
                        Histo(plus_bins,ind) =  Histo(plus_bins,ind) + factor_plus*Mag;
                        
                    else            % condition generale
                        
                        
                        factor_minus = (Bins_vector(plus_bins) - Dir) / Inter_bins;                     % quanty factor to put in the minus bins
                        factor_plus = (Dir - Bins_vector(minus_bins)) / Inter_bins;                      % quanty factor to put in the plus bins
                        
                        Histo(minus_bins,ind) =  Histo(minus_bins,ind) + factor_minus*Mag;
                        Histo(plus_bins,ind) =  Histo(plus_bins,ind) + factor_plus*Mag;
                        
                    end        
                end
            end
        end
        
        %_____________________________________________________________________________________________________________________________________________________
        
        % creat the Descriptor vector
        Des = [];
        for s=1:length(Histo(1,:))
            Des=[Des;Histo(:,s)];
        end
        
        % % normalisation of this vecxtor
        NORM = norm(Des);   
        if NORM==0
            Des = Des;
        else
            Des = Des / NORM;
            
%             % verification if there is an element bigger than 0.4
%             for t=1:length(Des)
%                 if Des(t)>(10/N_bins)*0.16              % 0.16 is a good value for 10 bins 
%                     Des(t)=(10/N_bins)*0.16;
%                 end
%             end    
%             
%             % re-normalisation of descriptor vector
%             NORM = norm(Des);
%             Des = Des / NORM ;  
            
        end
        %_____________________________________________________________________________________________________________________________________________________
        
        Pat(w).Descriptor(k).Descriptor = Des; 
    end
end

return

%--------------------------------------------------------------------------
%--------------------------------------------------------------------------

function ang = Modul_2pi(x)
for i=1:length(x)
    while x(i)>=pi,
        x(i)=x(i)-2*pi;
    end
    while x(i)<-pi,
        x(i)=x(i)+2*pi;
    end
    ang(i)=x(i);
end
return;