%--------------------------------------------------------------------------
% This program has been made to understand how to use my functions to
% extract and match vertical lines from omnidirectional images
% In this example the images belong to a stero omnidirectional camera
%
% Author: Davide Scaramuzza, June 2008
% Email: davide.scaramuzza@ieee.org
% If you use this code or re-implement the algorithm by yoursefl, please
% cite the publications:
%
% [1] Scaramuzza, D., Martinelli, A., and Siegwart, R. (2009)
% A Robust Descriptor for Tracking Vertical Lines in Omnidirectional Images and its Use in Mobile Robotics, 
% International Journal on Robotics Research, vol. 28, issue 2, February, 2009.
%
% [2] Scaramuzza, D., Criblez, N., Martinelli, A. and Siegwart, R. (2007), 
% Robust Feature Extraction and Matching for Omnidirectional Images, 
% Proceedings of the 6th International Conference on Field and Service Robotics, July, 2007
%--------------------------------------------------------------------------

clc;
clear all;
close all;

%--------------------------------------------------------------------------
% Initialize calib parameters of camera 1
%--------------------------------------------------------------------------
par1.number_of_pattern_for_one_feat = 3; % Number of circular areas
par1.Centre.X_centre = 248.5062;         % Coordinates of the image center (row)
par1.Centre.Y_centre = 325.1429;         % Coordinates of the image center (column)
par1.Rmin = 100;                         % Internal radius of the image (minimum distance from the center); the pixels inside will be cut out
par1.Rmax = 200;                         % External radius of the image (maximum distance from the center); the pixels outside will be cut out
par1.N_bins = 32;                        % Number of histogram bins to use for the descriptor
par1.radius = 256.3085;                  % Radius of the mirror
center1 = [par1.Centre.Y_centre, par1.Centre.X_centre];
par1.width = 640;
par1.height = 480;
%--------------------------------------------------------------------------
% Initialize calib parameters of the camera 2 (same as before but image
% center and radius are different
%--------------------------------------------------------------------------
par2 = par1;
par2.Centre.X_centre = 243.0395; %Coordinates of the image center
par2.Centre.Y_centre = 346.3323;
par2.radius = 254.5972;
center2 = [par2.Centre.Y_centre, par2.Centre.X_centre];
%--------------------------------------------------------------------------


%--------------------------------------------------------------------------
% Read Images
%--------------------------------------------------------------------------    
fprintf('Opening stereopair\n');

img1 = imread('img1.jpg');   %Image from the back camera

img2 = imread('img2.jpg');   %Image from the back camera

%Check if it is a grayscale or color image
if length(size(img1)) == 3
    imgg1 = rgb2gray(img1);
    imgg2 = rgb2gray(img2);
else
    imgg1 = img1;
    imgg2 = img2;
end

%--------------------------------------------------------------------------
% Extract Vertical Lines
%--------------------------------------------------------------------------
fprintf('Extracting vertical lines\n');
[verticals1,grad1] = vertical_features(double(imgg1), par1);
[verticals2,grad2] = vertical_features(double(imgg2), par2);

%--------------------------------------------------------------------------
% Visualize Verticals
%--------------------------------------------------------------------------
figure(1); 
subplot(1,2,1); imagesc(img1); title('Image 1');
draw2DCircle(center1,par1.radius,'r'); draw2DCircle(center1,par1.Rmax,'g'); draw2DCircle(center1,par1.Rmin,'b');
draw_verticals(verticals1.Vertical_Line, par1,'b'); drawnow;
if length(size(img1)) == 2, colormap(gray), end

%Draw circles on image 2
subplot(1,2,2); imagesc(img2); title('Image 2'); 
draw2DCircle(center2,par2.radius,'r'); draw2DCircle(center2,par2.Rmax,'g'); draw2DCircle(center2,par2.Rmin,'b');
draw_verticals(verticals2.Vertical_Line, par2,'b'); drawnow;
if length(size(img2)) == 2, colormap(gray), end

%--------------------------------------------------------------------------
% Extract Descriptors
%--------------------------------------------------------------------------
fprintf('Computing descrptors\n');
%des1 and des2 are NxM matrices, where N is the number of lines in the image and M the descriptor length
des1 = [];
des2 = [];
for ver = 1 : length(verticals1.Vertical_Line)
    des1(ver,:) = get_descriptor( par1 , grad1, verticals1.Vertical_Line(ver).Angle )'; %extract descriptor for each vertical
end
for ver = 1 : length(verticals2.Vertical_Line)
    des2(ver,:) = get_descriptor( par2 , grad2, verticals2.Vertical_Line(ver).Angle )'; %extract descriptor for each vertical
end


%--------------------------------------------------------------------------
% Match Verticals using Double-Consistency Check (it means that it matches from image 1 to 2 and from 2 to 1
%--------------------------------------------------------------------------
%TH is the threshold which is used to decide how similar are two
%descriptor, it is always < 1, try to augment TH to have more matched, but
%the higher TH, more prone is the result to flase positives


TH = 0.8;
fprintf('Matching descrptors\n');
[ind1, ind2] = match2check(des1, des2, TH); 
fprintf('Found %d LEFT/RIGHT matches.\n', length(ind1));

%--------------------------------------------------------------------------
% Visualize MATCHING Verticals
%--------------------------------------------------------------------------
%Draw circles on image 1
figure(2); 
subplot(1,2,1)
imagesc(img1);
title('Matching Lines Image 1');
draw2DCircle(center1,par1.radius,'r'); draw2DCircle(center1,par1.Rmax,'g'); draw2DCircle(center1,par1.Rmin,'b');
draw_verticals(verticals1.Vertical_Line(ind1), par1,'r'); drawnow;
if length(size(img1)) == 2, colormap(gray), end

%Draw circles on image 2
subplot(1,2,2)
imagesc(img2);
title('Matching Lines Image 2');
draw2DCircle(center2,par2.radius,'r'); draw2DCircle(center2,par2.Rmax,'g'); draw2DCircle(center2,par2.Rmin,'b');
draw_verticals(verticals2.Vertical_Line(ind2), par2,'r'); drawnow;
if length(size(img2)) == 2, colormap(gray), end



