% VERTICAL_FEATURES
%   VERTICAL_FEATURES returns the vertical lines of an omnidirectional
%   image
%   [VERTICALS,GRAD] = VERTICAL_FEATURES(I,  PARAMETERS, DISPLAY)
%   I is the input image
%   PARAMETERS is the parameters needed for extracting the lines, see the main program for an explication
%   DISPLAY. If not NULL it displaies the vertical lines overalid on the
%   original image, the figure number is given by the values of DISPLAY
%   GRAD is the image gradient
%   GRAD.PHASE is the gradient phase in radians. The phase is computed
%   according to the reference system displayed in Fig 6 of [1]
%   GRAd.MAG is the gradient magnitude
%   VERTICALS is a structure which contains all the vertical lines extracted
%   from a sequence; for an explication of the parameters see the main program
%   VERTICALS.VERTICAL_LINE.ANGLE is the orientation of each vertical
%   VERTICALS.VERTICAL_LINE.EDGES_NUMBER is the number of pixel of each vertical
%
% Author: Davide Scaramuzza, June 2008
% Email: davide.scaramuzza@ieee.org
% If you use this code or re-implement the algorithm by yoursefl, please
% cite the publications:
%
% [1] Scaramuzza, D., Martinelli, A., and Siegwart, R. (2009)
% A Robust Descriptor for Tracking Vertical Lines in Omnidirectional Images and its Use in Mobile Robotics, 
% International Journal on Robotics Research, vol. 28, issue 2, February, 2009.
%
% [2] Scaramuzza, D., Criblez, N., Martinelli, A. and Siegwart, R. (2007), 
% Robust Feature Extraction and Matching for Omnidirectional Images, 
% Proceedings of the 6th International Conference on Field and Service Robotics, July, 2007
%--------------------------------------------------------------------------

function [verticals,Gradient] = vertical_features(I, parameters, display)


%%% --------------------------------------- initialisation -------------------------------------------- %%%

if nargin <3
    display = 0;
end

b=I;

win = 3;                                                      % taille de la fenetre qui effecura le filtre mediant
Nmax = 720;                                                   % nombre max de vertical soit un max de une vertical tous les 1/2 deg                                                   
Rmin = parameters.Rmin;                                       
Rmax = parameters.Rmax;                                       
Nw = 3;                                                       %Taille de la fenetre de scanage
globalTh = 10;                                                % seuille de detection, si il y a plus de ponit qui ont le meme angle qui pointe vers zero que globalTh alors on deside que le feature est bon!
Min_angle_between_two_Vertical = 2;                           % in deg

centre_x = parameters.Centre.X_centre;           % coordonnee X du centre de l'image
centre_y = parameters.Centre.Y_centre;           % coordonnee Y du centre de l'image

%%% -----------------------------------------------------------------------

wx=[-1 -2 -1;0 0 0;1 2 1];                                  % Filtre pour la recherche du gradient horizontal (Sobel)
wy=wx';                                                     % Filtre pour la recherche du gradient vertical (Sobel)


[height,width]=size(b);                                     % ici b est l'image 

x=[1:height]';                                              
x=x*ones(1,width);                                          % x est maintenant une matrice de la meme taille que l'image, chaque ligne contient la valeur de la ligne en question soit il n'y a que des 8 dans la ligne 8.

y=[1:width];                                                
y=ones(height,1)*y;                                         % Y est maintenant une matrice de la meme taille que l'image, chaque colone contient la valeur de la colone en question soit il n'y a que des 8 dans la colone 8.              

x=x-centre_x;                                               % centrage
y=y-centre_y;

rho=sqrt(x.^2+y.^2);                                        % longuer entre le 0 et un point, rho est donc une matrice de la taille de l'image qui contient la distance entre chaque pixel et le centre de l'image! l'information est stocke dans le pixel en question

[l,c]=find(rho==0);                                         % juste pour eviter d'avoir un warning du a une division par zero lors du calkcul du cos et sin! la valeur n'imflue en rien le resultat car de toute facon elle seront abandonne par la suite car trop proche du centre
if isempty(find(rho==0))~=1
   [l,c]=find(rho==0);
   rho(l,c)=0.5;                            
end

theta=atan2(y,x);                                           % angle entre rho et l'axe X, theta est donc une matrice de la taille de l'image qui contient l'angle entre chaque pixel et l'axe des x! l'information est stocke dans le pixel en question
theta=round((theta+pi)/(2*pi)*Nmax);                        % rounds the elements of (theta+pi)/(2*pi)*Nmax to the nearest integers. 
costh=x./rho;                                               % cos=adj/hyp
sinth=y./rho;                                               % sin=opp/hyp
z=zeros(size(b));                                           % remplissage d'une image de 0 de la meme taille que les images
z(find(rho<Rmax & rho>Rmin))=1;                             % trouve toutes les longueur de rho qui sont plus longue et plus courte que Rmax et Rmin. Donc z contiendra des 1 dans l'interval souhait et de 0 ailleurs



%MAIN

b=b.*z;                                                         % on ne va garder que les pixel de l'image qui sont dans l'interval souhaite.
c = b; %c=medfilt2(b,[win,win]);                                        % medfilt2 function implements median filtering. 
grad_x=conv2(b,wx,'same');                                      % recherche du gradient de l'image C, ceci par convolution avec le filtre de Sobel (horizontal) 
grad_y=conv2(b,wy,'same');                                      % recherche du gradient de l'image C, ceci par convolution avec le filtre de Sobel (vertical) 


Grad_Theta = -sinth.*grad_x + costh.*grad_y;
dfnorm= abs(-sinth.*grad_x + costh.*grad_y);                    % dfnorm contiendra la valeur absolue de la derive selon l'angle theta qui ponit vers le centre
vertical=zeros(Nmax,1);                                         % initialisation du vecteur qui contiendra le nombre de ponit qui auront un gradient diriger vers le centre
dfnorm(find(dfnorm<0.05*max(max(dfnorm))))=0;                   % toutes les valeurs de dfnorm qui sont inferieur  5% de la valeur maximum de dfnorm sont mis a zeros
[ii,jj]=find(dfnorm);                                           % [i,j] = find(X) returns the row and column indices of the nonzero entries in the matrix X. This is often used with sparse matrices.
prova=zeros(size(dfnorm));                                      % matrice de zero de la taille des images

mag=sqrt(grad_x.^2+grad_y.^2);                                  % magnitude du gradient

Gradient.phase = atan2(grad_y,grad_x);                          % angle de chaque gradient (angle qui correspond  la valeur max du gradient pour le pixel considerer)
Gradient.mag = mag;                                            % angle de chaque gradient (angle qui correspond  la valeur max du gradient pour le pixel considerer)


for l=1:length(ii)
    i=ii(l);
    j=jj(l);
    
    iy=-y(i,j);
    ix=x(i,j);
    
    if i>1 & i<height & j>1 & j<width & ix~=0  
           if ((iy<=0 & ix>-iy)  | (iy>=0 & ix<-iy))
               d = abs(iy/ix);
               gradmag1 = mag(i,j+1).*(1-d) + mag(i-1,j+1).*d; 
               gradmag2 = mag(i,j-1).*(1-d) + mag(i+1,j-1).*d; 
               prova(i,j)=(dfnorm(i,j)>=gradmag1 & dfnorm(i,j)>=gradmag2); 
           elseif ((ix>0 & -iy>=ix)  | (ix<0 & -iy<=ix))
               d = abs(ix/iy); %abs(iy/ix);
               gradmag1 = mag(i-1,j).*(1-d) + mag(i-1,j+1).*d; 
               gradmag2 = mag(i+1,j).*(1-d) + mag(i+1,j-1).*d;    
               prova(i,j)=(dfnorm(i,j)>=gradmag1 & dfnorm(i,j)>=gradmag2); 
           elseif ((ix<=0 & ix>iy) | (ix>=0 & ix<iy))
               d = abs(ix/iy); %abs(iy/ix);
               gradmag1 = mag(i-1,j).*(1-d) + mag(i-1,j-1).*d; 
               gradmag2 = mag(i+1,j).*(1-d) + mag(i+1,j+1).*d;
               prova(i,j)=(dfnorm(i,j)>=gradmag1 & dfnorm(i,j)>=gradmag2); 
           elseif ((iy<0 & ix<=iy) | (iy>0 & ix>=iy))
               d = abs(iy/ix);
               gradmag1 = mag(i,j-1).*(1-d) + mag(i-1,j-1).*d; 
               gradmag2 = mag(i,j+1).*(1-d) + mag(i+1,j+1).*d;        
               prova(i,j)=(dfnorm(i,j)>=gradmag1 & dfnorm(i,j)>=gradmag2); 
           end
    end       

    if theta(i,j)==0
        continue;
    end
    
    if prova(i,j)==1
        index_angle = theta(i,j);
        temp_gradient(index_angle).Angle = Modul_2pi( theta(i,j)/Nmax*2*pi);
    end

    vertical(theta(i,j)) = vertical(theta(i,j))+prova(i,j);
    
end

ind = localmaximaglobal2(vertical,Min_angle_between_two_Vertical, globalTh); % localmaximaglobal(vertical,Nw,globalTh,Min_angle_between_two_Vertical);
i=(ind)./Nmax*2*pi;                                         % anisi i contiendra tous les angles en radian entre 0 et 2pi

% Non=maxima suppression
if isempty(ind)
    
    verticals.HaveVerticalLine_Yes_No = 0;        % 0 for no (have not vertical line)
else
    for index=1:length(ind)
        index_gradient = ind(index);
        verticals.Vertical_Line(index).Angle = temp_gradient(index_gradient).Angle;
    end
    for index=1:length(ind)
        index_gradient = ind(index);
        verticals.Vertical_Line(index).edges_number = vertical(index_gradient);     
    end    
    verticals.HaveVerticalLine_Yes_No = 1;        % 1 for yes (have vertical line)
end

%_____________________________________________________________________________________________________________________________________________________
% Plot the image and the vertical line
if display
    figure(display);
    imagesc(b)
    colormap(gray)
    drawline(b,centre_x,centre_y,i);
    drawnow;
    % figure(11); imagesc(prova); colormap(gray);
end
%_____________________________________________________________________________________________________________________________________________________


%_____________________________________________________________________________________________________________________________________________________
% Plot the number of point who correspond to a vertical line
% figure(10); imagesc(prova); colormap(gray);                             % image en noir et blanc qui contient juste les points qui on des gradient qui point vers le centre, soit qui font partie d'un feature vertical
% plot(vertical);                                             % dessin le "spectre" des lignes verticales
% hold on; 
% plot(ind,vertical(ind),'ro');                               % rajoute des petit rond rouge au sommet des pics
% line([0,length(vertical)],[globalTh,globalTh]);             % dessine une ligne va de 0 a length(vertical) en x et de globalTh a globalTh en y (soit une ligne horizontal a une hauteur definie par globalTh)
% xlabel('Angle on image [1/2 deg]'); 
% ylabel('Number of points that have a good gradient direction');
% hold off;
% axis([0 720 0 100]);                                        % l'axe des x ira de 0 a 720 et les y de 0 a 100
%_____________________________________________________________________________________________________________________________________________________


return